import json
import warnings
from typing import Any, cast

import litellm
from pydantic import BaseModel

from blackgeorge.adapters.base import BaseModelAdapter, ModelResponse
from blackgeorge.adapters.instructor_client import instructor_clients
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.utils import new_id

warnings.filterwarnings("ignore", message="Pydantic serializer warnings:.*", category=UserWarning)


def _get(obj: Any, key: str, default: Any = None) -> Any:
    if isinstance(obj, dict):
        return obj.get(key, default)
    return getattr(obj, key, default)


def _parse_tool_calls(message: Any) -> list[ToolCall]:
    tool_calls = _get(message, "tool_calls", []) or []
    parsed: list[ToolCall] = []

    for call in tool_calls:
        function = _get(call, "function", {})
        name = _get(function, "name")
        arguments_raw = _get(function, "arguments")
        arguments: dict[str, Any] = {}
        error: str | None = None

        if isinstance(arguments_raw, str) and arguments_raw:
            try:
                arguments = json.loads(arguments_raw)
            except json.JSONDecodeError as e:
                error = f"Invalid JSON in tool arguments: {e}. Raw: {arguments_raw[:100]}"
                arguments = {}

        call_id = _get(call, "id") or new_id()
        parsed.append(ToolCall(id=call_id, name=name, arguments=arguments, error=error))

    return parsed


def _parse_response(response: Any) -> ModelResponse:
    choices = _get(response, "choices", [])
    message = _get(choices[0], "message") if choices else None
    content = _get(message, "content") if message else None
    reasoning_content = _get(message, "reasoning_content") if message else None
    tool_calls = _parse_tool_calls(message) if message else []
    usage = _get(response, "usage", {}) or {}
    if isinstance(usage, BaseModel):
        usage = usage.model_dump(mode="json", warnings=False)
    return ModelResponse(
        content=content,
        reasoning_content=reasoning_content,
        tool_calls=tool_calls,
        usage=usage,
        raw=response,
    )


class LiteLLMAdapter(BaseModelAdapter):
    def complete(
        self,
        *,
        model: str,
        messages: list[dict[str, Any]],
        tools: list[dict[str, Any]] | None,
        tool_choice: str | dict[str, Any] | None,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        thinking: dict[str, Any] | None = None,
        drop_params: bool | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> ModelResponse | list[dict[str, Any]]:
        stream_options = stream_options if stream else None
        litellm_params: dict[str, Any] = {
            "model": model,
            "messages": messages,
            "tools": tools,
            "tool_choice": tool_choice,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "stream": stream,
            "stream_options": stream_options,
        }

        if thinking is not None:
            litellm_params["thinking"] = thinking
        if drop_params is not None:
            litellm_params["drop_params"] = drop_params
        if extra_body is not None:
            litellm_params["extra_body"] = extra_body

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Pydantic serializer warnings:.*")
            response = litellm.completion(**litellm_params)
        if stream:
            return cast(list[dict[str, Any]], response)
        return _parse_response(response)

    async def acomplete(
        self,
        *,
        model: str,
        messages: list[dict[str, Any]],
        tools: list[dict[str, Any]] | None,
        tool_choice: str | dict[str, Any] | None,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        thinking: dict[str, Any] | None = None,
        drop_params: bool | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> ModelResponse | Any:
        stream_options = stream_options if stream else None
        litellm_params: dict[str, Any] = {
            "model": model,
            "messages": messages,
            "tools": tools,
            "tool_choice": tool_choice,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "stream": stream,
            "stream_options": stream_options,
        }

        if thinking is not None:
            litellm_params["thinking"] = thinking
        if drop_params is not None:
            litellm_params["drop_params"] = drop_params
        if extra_body is not None:
            litellm_params["extra_body"] = extra_body

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Pydantic serializer warnings:.*")
            response = await litellm.acompletion(**litellm_params)
        if stream:
            return response
        return _parse_response(response)

    def structured_complete(
        self,
        *,
        model: str,
        messages: list[dict[str, Any]],
        response_schema: Any,
        retries: int,
    ) -> Any:
        payload = list(messages)
        client = instructor_clients.get(model, async_client=False)
        attempts = 0
        while True:
            try:
                return client.chat.completions.create(
                    model=model,
                    messages=payload,
                    response_model=response_schema,
                )
            except Exception as exc:
                if attempts >= retries:
                    raise
                payload.append(
                    {
                        "role": "user",
                        "content": f"Fix validation errors: {exc}",
                    }
                )
                attempts += 1

    async def astructured_complete(
        self,
        *,
        model: str,
        messages: list[dict[str, Any]],
        response_schema: Any,
        retries: int,
    ) -> Any:
        payload = list(messages)
        client = instructor_clients.get(model, async_client=True)
        attempts = 0
        while True:
            try:
                return await client.chat.completions.create(
                    model=model,
                    messages=payload,
                    response_model=response_schema,
                )
            except Exception as exc:
                if attempts >= retries:
                    raise
                payload.append(
                    {
                        "role": "user",
                        "content": f"Fix validation errors: {exc}",
                    }
                )
                attempts += 1



import json
from collections.abc import Callable, Iterable
from typing import Any, cast

from blackgeorge.adapters.base import BaseModelAdapter, ModelResponse
from blackgeorge.adapters.instructor_client import instructor_clients
from blackgeorge.core.event import Event
from blackgeorge.core.job import Job
from blackgeorge.core.message import Message
from blackgeorge.core.pending_action import PendingAction
from blackgeorge.core.report import Report
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.core.types import RunStatus
from blackgeorge.store.state import RunState
from blackgeorge.tools.base import Tool, ToolResult
from blackgeorge.tools.execution import aexecute_tool, execute_tool
from blackgeorge.tools.registry import Toolbelt
from blackgeorge.utils import new_id
from blackgeorge.worker_context import (
    CONTEXT_SUMMARY_MAX_ATTEMPTS,
    aapply_context_summary,
    apply_context_summary,
    context_error_message,
    is_context_limit_error,
    litellm_model_registered,
)
from blackgeorge.worker_messages import (
    chunk_content,
    chunk_usage,
    emit_assistant_message,
    ensure_content,
    messages_to_payload,
    render_input,
    replace_tool_call,
    structured_content,
    system_message,
    tool_call_with_result,
    tool_message,
    tool_schemas,
)
from blackgeorge.worker_tools import (
    pending_options,
    resume_argument_key,
    tool_action_type,
    tool_prompt,
    update_arguments,
)

EventEmitter = Callable[[str, str, dict[str, Any]], None]


def _build_report(
    run_id: str,
    status: RunStatus,
    content: str | None = None,
    reasoning_content: str | None = None,
    data: Any | None = None,
    messages: list[Message] | None = None,
    tool_calls: list[ToolCall] | None = None,
    metrics: dict[str, Any] | None = None,
    events: list[Event] | None = None,
    pending_action: PendingAction | None = None,
    errors: list[str] | None = None,
) -> Report:
    return Report(
        run_id=run_id,
        status=status,
        content=content,
        reasoning_content=reasoning_content,
        data=data,
        messages=list(messages) if messages else [],
        tool_calls=list(tool_calls) if tool_calls else [],
        metrics=metrics or {},
        events=list(events) if events else [],
        pending_action=pending_action,
        errors=list(errors) if errors else [],
    )


def _build_state(
    run_id: str,
    status: RunStatus,
    runner_name: str,
    job: Job,
    messages: list[Message],
    tool_calls: list[ToolCall],
    pending_action: PendingAction | None,
    metrics: dict[str, Any],
    iteration: int,
    payload: dict[str, Any] | None = None,
) -> RunState:
    return RunState(
        run_id=run_id,
        status=status,
        runner_type="worker",
        runner_name=runner_name,
        job=job,
        messages=messages,
        tool_calls=tool_calls,
        pending_action=pending_action,
        metrics=metrics,
        iteration=iteration,
        payload=payload or {},
    )


def _report_error(
    run_id: str,
    messages: list[Message],
    errors: list[str],
    events: list[Event],
) -> Report:
    return Report(
        run_id=run_id,
        status="failed",
        content=None,
        reasoning_content=None,
        data=None,
        messages=messages,
        tool_calls=[],
        metrics={},
        events=events,
        pending_action=None,
        errors=errors,
    )


def _should_stream(stream: bool, tools: list[Tool], response_schema: Any | None) -> bool:
    return stream and not tools and response_schema is None


def _tool_result_preview(result: ToolResult, limit: int) -> tuple[str | None, bool]:
    if result.content is not None:
        text = result.content
    elif result.data is not None:
        try:
            text = json.dumps(result.data, ensure_ascii=True)
        except (TypeError, ValueError):
            text = str(result.data)
    elif result.error is not None:
        text = result.error
    else:
        return None, False
    if len(text) > limit:
        return f"{text[:limit]}...", True
    return text, False


def _tool_event_payload(call: ToolCall, result: ToolResult, limit: int = 200) -> dict[str, Any]:
    payload: dict[str, Any] = {"tool_call_id": call.id}
    preview, truncated = _tool_result_preview(result, limit)
    if preview is not None:
        payload["result_preview"] = preview
        payload["result_truncated"] = truncated
    if result.timed_out:
        payload["timed_out"] = True
    if result.cancelled:
        payload["cancelled"] = True
    return payload


class WorkerRunner:
    def __init__(self, name: str, toolbelt: Toolbelt, instructions: str | None) -> None:
        self.name = name
        self.toolbelt = toolbelt
        self.instructions = instructions

    def _build_messages(self, job: Job) -> list[Message]:
        messages: list[Message] = []

        if job.initial_messages:
            messages.extend(job.initial_messages)

        system_content = system_message(self.instructions, job)
        if system_content:
            if not messages or messages[0].role != "system":
                messages.insert(0, Message(role="system", content=system_content))
            else:
                messages[0] = Message(
                    role="system",
                    content=f"{messages[0].content}\n\n{system_content}",
                )

        messages.append(Message(role="user", content=render_input(job.input)))
        return messages

    def _resolve_tools(self, job: Job) -> list[Tool]:
        if job.tools_override is not None:
            resolved: list[Tool] = []
            for item in job.tools_override:
                if isinstance(item, Tool):
                    resolved.append(item)
                    continue
                if isinstance(item, str):
                    tool = self.toolbelt.resolve(item)
                    if tool is not None:
                        resolved.append(tool)
            return resolved
        return self.toolbelt.list()

    def _structured_completion(
        self,
        *,
        adapter: BaseModelAdapter,
        model: str,
        messages: list[Message],
        response_schema: Any,
        retries: int,
    ) -> Any:
        payload = messages_to_payload(messages)
        try:
            return adapter.structured_complete(
                model=model,
                messages=payload,
                response_schema=response_schema,
                retries=retries,
            )
        except NotImplementedError:
            pass
        client = instructor_clients.get(model, async_client=False)
        attempts = 0
        while True:
            try:
                return client.chat.completions.create(
                    model=model,
                    messages=payload,
                    response_model=response_schema,
                )
            except Exception as exc:
                if attempts >= retries:
                    raise
                payload.append(
                    {
                        "role": "user",
                        "content": f"Fix validation errors: {exc}",
                    }
                )
                attempts += 1

    async def _astructured_completion(
        self,
        *,
        adapter: BaseModelAdapter,
        model: str,
        messages: list[Message],
        response_schema: Any,
        retries: int,
    ) -> Any:
        payload = messages_to_payload(messages)
        try:
            return await adapter.astructured_complete(
                model=model,
                messages=payload,
                response_schema=response_schema,
                retries=retries,
            )
        except NotImplementedError:
            pass
        client = instructor_clients.get(model, async_client=True)
        attempts = 0
        while True:
            try:
                return await client.chat.completions.create(
                    model=model,
                    messages=payload,
                    response_model=response_schema,
                )
            except Exception as exc:
                if attempts >= retries:
                    raise
                payload.append(
                    {
                        "role": "user",
                        "content": f"Fix validation errors: {exc}",
                    }
                )
                attempts += 1

    def _completion(
        self,
        *,
        adapter: BaseModelAdapter,
        model: str,
        messages: list[Message],
        tools: list[Tool],
        temperature: float | None,
        max_tokens: int | None,
        stream_options: dict[str, Any] | None,
        thinking: dict[str, Any] | None = None,
        drop_params: bool | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> ModelResponse:
        response = adapter.complete(
            model=model,
            messages=messages_to_payload(messages),
            tools=tool_schemas(tools) if tools else None,
            tool_choice="auto" if tools else None,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=False,
            stream_options=stream_options,
            thinking=thinking,
            drop_params=drop_params,
            extra_body=extra_body,
        )
        if isinstance(response, ModelResponse):
            return response
        return ModelResponse(content=None, tool_calls=[], usage={}, raw=response)

    async def _acompletion(
        self,
        *,
        adapter: BaseModelAdapter,
        model: str,
        messages: list[Message],
        tools: list[Tool],
        temperature: float | None,
        max_tokens: int | None,
        stream_options: dict[str, Any] | None,
        thinking: dict[str, Any] | None = None,
        drop_params: bool | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> ModelResponse:
        response = await adapter.acomplete(
            model=model,
            messages=messages_to_payload(messages),
            tools=tool_schemas(tools) if tools else None,
            tool_choice="auto" if tools else None,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=False,
            stream_options=stream_options,
            thinking=thinking,
            drop_params=drop_params,
            extra_body=extra_body,
        )
        if isinstance(response, ModelResponse):
            return response
        return ModelResponse(content=None, tool_calls=[], usage={}, raw=response)

    def _stream_completion(
        self,
        *,
        adapter: BaseModelAdapter,
        model: str,
        messages: list[Message],
        temperature: float | None,
        max_tokens: int | None,
        stream_options: dict[str, Any] | None,
        on_token: Callable[[str], None],
        thinking: dict[str, Any] | None = None,
        drop_params: bool | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> ModelResponse:
        stream = cast(
            Iterable[Any],
            adapter.complete(
                model=model,
                messages=messages_to_payload(messages),
                tools=None,
                tool_choice=None,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=True,
                stream_options=stream_options,
                thinking=thinking,
                drop_params=drop_params,
                extra_body=extra_body,
            ),
        )
        content_parts: list[str] = []
        usage: dict[str, Any] = {}
        for chunk in stream:
            token = chunk_content(chunk)
            if token:
                content_parts.append(token)
                on_token(token)
            usage_chunk = chunk_usage(chunk)
            if usage_chunk:
                usage = usage_chunk
        return ModelResponse(
            content="".join(content_parts),
            tool_calls=[],
            usage=usage,
            raw=stream,
        )

    async def _astream_completion(
        self,
        *,
        adapter: BaseModelAdapter,
        model: str,
        messages: list[Message],
        temperature: float | None,
        max_tokens: int | None,
        stream_options: dict[str, Any] | None,
        on_token: Callable[[str], None],
        thinking: dict[str, Any] | None = None,
        drop_params: bool | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> ModelResponse:
        stream = await adapter.acomplete(
            model=model,
            messages=messages_to_payload(messages),
            tools=None,
            tool_choice=None,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=True,
            stream_options=stream_options,
            thinking=thinking,
            drop_params=drop_params,
            extra_body=extra_body,
        )
        content_parts: list[str] = []
        usage: dict[str, Any] = {}
        if hasattr(stream, "__aiter__"):
            async for chunk in stream:
                token = chunk_content(chunk)
                if token:
                    content_parts.append(token)
                    on_token(token)
                usage_chunk = chunk_usage(chunk)
                if usage_chunk:
                    usage = usage_chunk
        else:
            for chunk in cast(Iterable[Any], stream):
                token = chunk_content(chunk)
                if token:
                    content_parts.append(token)
                    on_token(token)
                usage_chunk = chunk_usage(chunk)
                if usage_chunk:
                    usage = usage_chunk
        return ModelResponse(
            content="".join(content_parts),
            tool_calls=[],
            usage=usage,
            raw=stream,
        )

    def _run_loop(
        self,
        *,
        adapter: BaseModelAdapter,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        messages: list[Message],
        tool_calls: list[ToolCall],
        metrics: dict[str, Any],
        errors: list[str],
        iteration: int,
        model_name: str,
        respect_context_window: bool,
    ) -> tuple[Report, RunState | None]:
        tools = self._resolve_tools(job)
        allowed_tools = {tool.name: tool for tool in tools}
        response_schema = job.response_schema
        context_summaries = 0
        model_registered = litellm_model_registered(model_name)

        while iteration < max_iterations:
            iteration += 1
            if _should_stream(stream, tools, response_schema):
                try:
                    response = self._stream_completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stream_options=stream_options,
                        on_token=lambda token: emit("stream.token", self.name, {"token": token}),
                        thinking=job.thinking,
                        drop_params=job.drop_params,
                        extra_body=job.extra_body,
                    )
                except Exception as exc:
                    if not is_context_limit_error(exc):
                        raise
                    if not respect_context_window:
                        errors.append(context_error_message(model_registered, False))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    if context_summaries >= CONTEXT_SUMMARY_MAX_ATTEMPTS:
                        errors.append(context_error_message(model_registered, True))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    if not apply_context_summary(
                        adapter=adapter,
                        model_name=model_name,
                        messages=messages,
                        temperature=temperature,
                        metrics=metrics,
                        emit=emit,
                        worker_name=self.name,
                        model_registered=model_registered,
                    ):
                        errors.append(context_error_message(model_registered, True))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    context_summaries += 1
                    continue
            elif response_schema is not None and not tools:
                try:
                    data = self._structured_completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        response_schema=response_schema,
                        retries=structured_output_retries,
                    )
                except Exception as exc:
                    if is_context_limit_error(exc):
                        if not respect_context_window:
                            errors.append(context_error_message(model_registered, False))
                            emit("worker.failed", self.name, {"error": errors[-1]})
                            report = _build_report(
                                run_id,
                                "failed",
                                None,
                                None,
                                None,
                                messages,
                                tool_calls,
                                metrics,
                                events,
                                None,
                                errors,
                            )
                            return report, None
                        if context_summaries >= CONTEXT_SUMMARY_MAX_ATTEMPTS:
                            errors.append(context_error_message(model_registered, True))
                            emit("worker.failed", self.name, {"error": errors[-1]})
                            report = _build_report(
                                run_id,
                                "failed",
                                None,
                                None,
                                None,
                                messages,
                                tool_calls,
                                metrics,
                                events,
                                None,
                                errors,
                            )
                            return report, None
                        if not apply_context_summary(
                            adapter=adapter,
                            model_name=model_name,
                            messages=messages,
                            temperature=temperature,
                            metrics=metrics,
                            emit=emit,
                            worker_name=self.name,
                            model_registered=model_registered,
                        ):
                            errors.append(context_error_message(model_registered, True))
                            emit("worker.failed", self.name, {"error": errors[-1]})
                            report = _build_report(
                                run_id,
                                "failed",
                                None,
                                None,
                                None,
                                messages,
                                tool_calls,
                                metrics,
                                events,
                                None,
                                errors,
                            )
                            return report, None
                        context_summaries += 1
                        continue
                    errors.append(str(exc))
                    emit("worker.failed", self.name, {"error": errors[-1]})
                    report = _build_report(
                        run_id,
                        "failed",
                        None,
                        None,
                        None,
                        messages,
                        tool_calls,
                        metrics,
                        events,
                        None,
                        errors,
                    )
                    return report, None
                content = structured_content(data)
                assistant_message = Message(role="assistant", content=content)
                messages.append(assistant_message)
                emit_assistant_message(emit, self.name, assistant_message)
                emit("worker.completed", self.name, {})
                report = _build_report(
                    run_id,
                    "completed",
                    content,
                    None,
                    data,
                    messages,
                    tool_calls,
                    metrics,
                    events,
                    None,
                    errors,
                )
                return report, None
            else:
                try:
                    response = self._completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        tools=tools,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stream_options=stream_options,
                        thinking=job.thinking,
                        drop_params=job.drop_params,
                        extra_body=job.extra_body,
                    )
                except Exception as exc:
                    if not is_context_limit_error(exc):
                        raise
                    if not respect_context_window:
                        errors.append(context_error_message(model_registered, False))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    if context_summaries >= CONTEXT_SUMMARY_MAX_ATTEMPTS:
                        errors.append(context_error_message(model_registered, True))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    if not apply_context_summary(
                        adapter=adapter,
                        model_name=model_name,
                        messages=messages,
                        temperature=temperature,
                        metrics=metrics,
                        emit=emit,
                        worker_name=self.name,
                        model_registered=model_registered,
                    ):
                        errors.append(context_error_message(model_registered, True))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    context_summaries += 1
                    continue

            metrics["usage"] = response.usage if isinstance(response, ModelResponse) else {}

            if response.tool_calls:
                assistant_message = Message(
                    role="assistant",
                    content=ensure_content(response.content),
                    tool_calls=response.tool_calls,
                )
                messages.append(assistant_message)
                emit_assistant_message(emit, self.name, assistant_message)
                for call in response.tool_calls:
                    tool_calls.append(call)
                    if len(tool_calls) >= max_tool_calls:
                        errors.append("Max tool calls exceeded")
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    if call.error:
                        result = ToolResult(error=call.error)
                        emit(
                            "tool.failed",
                            call.name,
                            {"tool_call_id": call.id, "error": result.error},
                        )
                        messages.append(tool_message(result, call))
                        replace_tool_call(tool_calls, tool_call_with_result(call, result))
                        continue

                    tool = allowed_tools.get(call.name)
                    if tool is None:
                        result = ToolResult(error=f"Tool not found: {call.name}")
                        emit(
                            "tool.failed",
                            call.name,
                            {"tool_call_id": call.id, "error": result.error},
                        )
                        messages.append(tool_message(result, call))
                        replace_tool_call(tool_calls, tool_call_with_result(call, result))
                        continue
                    action_type = tool_action_type(tool)
                    if action_type:
                        metadata = {"tool": tool.name}
                        if tool.input_key:
                            metadata["input_key"] = tool.input_key
                        pending = PendingAction(
                            action_id=new_id(),
                            type=action_type,
                            tool_call=call,
                            prompt=tool_prompt(tool, action_type, call),
                            options=pending_options(action_type),
                            metadata=metadata,
                        )
                        emit(
                            f"tool.{action_type}_requested",
                            tool.name,
                            {"tool_call_id": call.id},
                        )
                        emit("worker.completed", self.name, {})
                        report = _build_report(
                            run_id,
                            "paused",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            pending,
                            errors,
                        )
                        state = _build_state(
                            run_id,
                            "paused",
                            self.name,
                            job,
                            messages,
                            tool_calls,
                            pending,
                            metrics,
                            iteration,
                        )
                        return report, state

                    emit("tool.started", tool.name, {"tool_call_id": call.id})
                    result = execute_tool(tool, call)
                    if result.error:
                        emit(
                            "tool.failed",
                            tool.name,
                            {"tool_call_id": call.id, "error": result.error},
                        )
                    else:
                        emit("tool.completed", tool.name, _tool_event_payload(call, result))
                    tool_result_message = tool_message(result, call)
                    messages.append(tool_result_message)
                    replace_tool_call(tool_calls, tool_call_with_result(call, result))
                continue

            if response_schema is not None:
                try:
                    data = self._structured_completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        response_schema=response_schema,
                        retries=structured_output_retries,
                    )
                except Exception as exc:
                    errors.append(str(exc))
                    emit("worker.failed", self.name, {"error": errors[-1]})
                    report = _build_report(
                        run_id,
                        "failed",
                        None,
                        None,
                        None,
                        messages,
                        tool_calls,
                        metrics,
                        events,
                        None,
                        errors,
                    )
                    return report, None
                content = structured_content(data)
                assistant_message = Message(role="assistant", content=content)
                messages.append(assistant_message)
                emit_assistant_message(emit, self.name, assistant_message)
                emit("worker.completed", self.name, {})
                report = _build_report(
                    run_id,
                    "completed",
                    content,
                    None,
                    data,
                    messages,
                    tool_calls,
                    metrics,
                    events,
                    None,
                    errors,
                )
                return report, None

            assistant_message = Message(role="assistant", content=response.content or "")
            messages.append(assistant_message)
            emit_assistant_message(emit, self.name, assistant_message)
            emit("worker.completed", self.name, {})
            report = _build_report(
                run_id,
                "completed",
                response.content,
                response.reasoning_content,
                None,
                messages,
                tool_calls,
                metrics,
                events,
                None,
                errors,
            )
            return report, None

        errors.append("Max iterations exceeded")
        emit("worker.failed", self.name, {"error": errors[-1]})
        report = _build_report(
            run_id,
            "failed",
            None,
            None,
            None,
            messages,
            tool_calls,
            metrics,
            events,
            None,
            errors,
        )
        return report, None

    async def _arun_loop(
        self,
        *,
        adapter: BaseModelAdapter,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        messages: list[Message],
        tool_calls: list[ToolCall],
        metrics: dict[str, Any],
        errors: list[str],
        iteration: int,
        model_name: str,
        respect_context_window: bool,
    ) -> tuple[Report, RunState | None]:
        tools = self._resolve_tools(job)
        allowed_tools = {tool.name: tool for tool in tools}
        response_schema = job.response_schema
        context_summaries = 0
        model_registered = litellm_model_registered(model_name)

        while iteration < max_iterations:
            iteration += 1
            if _should_stream(stream, tools, response_schema):
                try:
                    response = await self._astream_completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stream_options=stream_options,
                        on_token=lambda token: emit("stream.token", self.name, {"token": token}),
                        thinking=job.thinking,
                        drop_params=job.drop_params,
                        extra_body=job.extra_body,
                    )
                except Exception as exc:
                    if not is_context_limit_error(exc):
                        raise
                    if not respect_context_window:
                        errors.append(context_error_message(model_registered, False))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    if context_summaries >= CONTEXT_SUMMARY_MAX_ATTEMPTS:
                        errors.append(context_error_message(model_registered, True))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    if not await aapply_context_summary(
                        adapter=adapter,
                        model_name=model_name,
                        messages=messages,
                        temperature=temperature,
                        metrics=metrics,
                        emit=emit,
                        worker_name=self.name,
                        model_registered=model_registered,
                    ):
                        errors.append(context_error_message(model_registered, True))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    context_summaries += 1
                    continue
            elif response_schema is not None and not tools:
                try:
                    data = await self._astructured_completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        response_schema=response_schema,
                        retries=structured_output_retries,
                    )
                except Exception as exc:
                    if is_context_limit_error(exc):
                        if not respect_context_window:
                            errors.append(context_error_message(model_registered, False))
                            emit("worker.failed", self.name, {"error": errors[-1]})
                            report = _build_report(
                                run_id,
                                "failed",
                                None,
                                None,
                                None,
                                messages,
                                tool_calls,
                                metrics,
                                events,
                                None,
                                errors,
                            )
                            return report, None
                        if context_summaries >= CONTEXT_SUMMARY_MAX_ATTEMPTS:
                            errors.append(context_error_message(model_registered, True))
                            emit("worker.failed", self.name, {"error": errors[-1]})
                            report = _build_report(
                                run_id,
                                "failed",
                                None,
                                None,
                                None,
                                messages,
                                tool_calls,
                                metrics,
                                events,
                                None,
                                errors,
                            )
                            return report, None
                        if not await aapply_context_summary(
                            adapter=adapter,
                            model_name=model_name,
                            messages=messages,
                            temperature=temperature,
                            metrics=metrics,
                            emit=emit,
                            worker_name=self.name,
                            model_registered=model_registered,
                        ):
                            errors.append(context_error_message(model_registered, True))
                            emit("worker.failed", self.name, {"error": errors[-1]})
                            report = _build_report(
                                run_id,
                                "failed",
                                None,
                                None,
                                None,
                                messages,
                                tool_calls,
                                metrics,
                                events,
                                None,
                                errors,
                            )
                            return report, None
                        context_summaries += 1
                        continue
                    errors.append(str(exc))
                    emit("worker.failed", self.name, {"error": errors[-1]})
                    report = _build_report(
                        run_id,
                        "failed",
                        None,
                        None,
                        None,
                        messages,
                        tool_calls,
                        metrics,
                        events,
                        None,
                        errors,
                    )
                    return report, None
                content = structured_content(data)
                assistant_message = Message(role="assistant", content=content)
                messages.append(assistant_message)
                emit_assistant_message(emit, self.name, assistant_message)
                emit("worker.completed", self.name, {})
                report = _build_report(
                    run_id,
                    "completed",
                    content,
                    None,
                    data,
                    messages,
                    tool_calls,
                    metrics,
                    events,
                    None,
                    errors,
                )
                return report, None
            else:
                try:
                    response = await self._acompletion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        tools=tools,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stream_options=stream_options,
                        thinking=job.thinking,
                        drop_params=job.drop_params,
                        extra_body=job.extra_body,
                    )
                except Exception as exc:
                    if not is_context_limit_error(exc):
                        raise
                    if not respect_context_window:
                        errors.append(context_error_message(model_registered, False))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    if context_summaries >= CONTEXT_SUMMARY_MAX_ATTEMPTS:
                        errors.append(context_error_message(model_registered, True))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    if not await aapply_context_summary(
                        adapter=adapter,
                        model_name=model_name,
                        messages=messages,
                        temperature=temperature,
                        metrics=metrics,
                        emit=emit,
                        worker_name=self.name,
                        model_registered=model_registered,
                    ):
                        errors.append(context_error_message(model_registered, True))
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    context_summaries += 1
                    continue

            metrics["usage"] = response.usage if isinstance(response, ModelResponse) else {}

            if response.tool_calls:
                assistant_message = Message(
                    role="assistant",
                    content=ensure_content(response.content),
                    tool_calls=response.tool_calls,
                )
                messages.append(assistant_message)
                emit_assistant_message(emit, self.name, assistant_message)
                for call in response.tool_calls:
                    tool_calls.append(call)
                    if len(tool_calls) >= max_tool_calls:
                        errors.append("Max tool calls exceeded")
                        emit("worker.failed", self.name, {"error": errors[-1]})
                        report = _build_report(
                            run_id,
                            "failed",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            None,
                            errors,
                        )
                        return report, None
                    if call.error:
                        result = ToolResult(error=call.error)
                        emit(
                            "tool.failed",
                            call.name,
                            {"tool_call_id": call.id, "error": result.error},
                        )
                        messages.append(tool_message(result, call))
                        replace_tool_call(tool_calls, tool_call_with_result(call, result))
                        continue

                    tool = allowed_tools.get(call.name)
                    if tool is None:
                        result = ToolResult(error=f"Tool not found: {call.name}")
                        emit(
                            "tool.failed",
                            call.name,
                            {"tool_call_id": call.id, "error": result.error},
                        )
                        messages.append(tool_message(result, call))
                        replace_tool_call(tool_calls, tool_call_with_result(call, result))
                        continue
                    action_type = tool_action_type(tool)
                    if action_type:
                        metadata = {"tool": tool.name}
                        if tool.input_key:
                            metadata["input_key"] = tool.input_key
                        pending = PendingAction(
                            action_id=new_id(),
                            type=action_type,
                            tool_call=call,
                            prompt=tool_prompt(tool, action_type, call),
                            options=pending_options(action_type),
                            metadata=metadata,
                        )
                        emit(
                            f"tool.{action_type}_requested",
                            tool.name,
                            {"tool_call_id": call.id},
                        )
                        emit("worker.completed", self.name, {})
                        report = _build_report(
                            run_id,
                            "paused",
                            None,
                            None,
                            None,
                            messages,
                            tool_calls,
                            metrics,
                            events,
                            pending,
                            errors,
                        )
                        state = _build_state(
                            run_id,
                            "paused",
                            self.name,
                            job,
                            messages,
                            tool_calls,
                            pending,
                            metrics,
                            iteration,
                        )
                        return report, state

                    emit("tool.started", tool.name, {"tool_call_id": call.id})
                    result = await aexecute_tool(tool, call)
                    if result.error:
                        emit(
                            "tool.failed",
                            tool.name,
                            {"tool_call_id": call.id, "error": result.error},
                        )
                    else:
                        emit("tool.completed", tool.name, _tool_event_payload(call, result))
                    tool_result_message = tool_message(result, call)
                    messages.append(tool_result_message)
                    replace_tool_call(tool_calls, tool_call_with_result(call, result))
                continue

            if response_schema is not None:
                try:
                    data = await self._astructured_completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        response_schema=response_schema,
                        retries=structured_output_retries,
                    )
                except Exception as exc:
                    errors.append(str(exc))
                    emit("worker.failed", self.name, {"error": errors[-1]})
                    report = _build_report(
                        run_id,
                        "failed",
                        None,
                        None,
                        None,
                        messages,
                        tool_calls,
                        metrics,
                        events,
                        None,
                        errors,
                    )
                    return report, None
                content = structured_content(data)
                assistant_message = Message(role="assistant", content=content)
                messages.append(assistant_message)
                emit_assistant_message(emit, self.name, assistant_message)
                emit("worker.completed", self.name, {})
                report = _build_report(
                    run_id,
                    "completed",
                    content,
                    None,
                    data,
                    messages,
                    tool_calls,
                    metrics,
                    events,
                    None,
                    errors,
                )
                return report, None

            assistant_message = Message(role="assistant", content=response.content or "")
            messages.append(assistant_message)
            emit_assistant_message(emit, self.name, assistant_message)
            emit("worker.completed", self.name, {})
            report = _build_report(
                run_id,
                "completed",
                response.content,
                response.reasoning_content,
                None,
                messages,
                tool_calls,
                metrics,
                events,
                None,
                errors,
            )
            return report, None

        errors.append("Max iterations exceeded")
        emit("worker.failed", self.name, {"error": errors[-1]})
        report = _build_report(
            run_id,
            "failed",
            None,
            None,
            None,
            messages,
            tool_calls,
            metrics,
            events,
            None,
            errors,
        )
        return report, None

    def run(
        self,
        *,
        adapter: BaseModelAdapter,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        messages = self._build_messages(job)
        tool_calls: list[ToolCall] = []
        metrics: dict[str, Any] = {}
        errors: list[str] = []
        iteration = 0

        if not model_name:
            errors.append("Worker model not set")
            emit("worker.failed", self.name, {"error": errors[-1]})
            report = _report_error(run_id, messages, errors, events)
            return report, None

        emit("worker.started", self.name, {})
        return self._run_loop(
            adapter=adapter,
            job=job,
            run_id=run_id,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            messages=messages,
            tool_calls=tool_calls,
            metrics=metrics,
            errors=errors,
            iteration=iteration,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    async def arun(
        self,
        *,
        adapter: BaseModelAdapter,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        messages = self._build_messages(job)
        tool_calls: list[ToolCall] = []
        metrics: dict[str, Any] = {}
        errors: list[str] = []
        iteration = 0

        if not model_name:
            errors.append("Worker model not set")
            emit("worker.failed", self.name, {"error": errors[-1]})
            report = _report_error(run_id, messages, errors, events)
            return report, None

        emit("worker.started", self.name, {})
        return await self._arun_loop(
            adapter=adapter,
            job=job,
            run_id=run_id,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            messages=messages,
            tool_calls=tool_calls,
            metrics=metrics,
            errors=errors,
            iteration=iteration,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    def resume(
        self,
        *,
        adapter: BaseModelAdapter,
        state: RunState,
        decision_or_input: Any,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        pending = state.pending_action
        if pending is None:
            report = _build_report(
                state.run_id,
                "failed",
                None,
                None,
                None,
                state.messages,
                state.tool_calls,
                state.metrics,
                events,
                None,
                ["No pending action"],
            )
            return report, None

        messages = list(state.messages)
        tool_calls = list(state.tool_calls)
        iteration = state.iteration
        metrics = dict(state.metrics)
        errors: list[str] = []

        tool = self.toolbelt.resolve(pending.tool_call.name)
        if tool is None:
            result = ToolResult(error=f"Tool not found: {pending.tool_call.name}")
            emit(
                "tool.failed",
                pending.tool_call.name,
                {"tool_call_id": pending.tool_call.id, "error": result.error},
            )
            messages.append(tool_message(result, pending.tool_call))
            replace_tool_call(tool_calls, tool_call_with_result(pending.tool_call, result))
        else:
            if pending.type == "confirmation" and not decision_or_input:
                result = ToolResult(error="Tool execution declined")
                tool_result_message = tool_message(result, pending.tool_call)
                messages.append(tool_result_message)
                replace_tool_call(tool_calls, tool_call_with_result(pending.tool_call, result))
            else:
                call = pending.tool_call
                if pending.type == "user_input":
                    key = resume_argument_key(pending)
                    call = update_arguments(call, key, decision_or_input)
                emit("tool.started", tool.name, {"tool_call_id": call.id})
                result = execute_tool(tool, call)
                if result.error:
                    emit("tool.failed", tool.name, {"tool_call_id": call.id, "error": result.error})
                else:
                    emit("tool.completed", tool.name, _tool_event_payload(call, result))
                tool_result_message = tool_message(result, call)
                messages.append(tool_result_message)
                replace_tool_call(tool_calls, tool_call_with_result(call, result))

        if not model_name:
            errors.append("Worker model not set")
            emit("worker.failed", self.name, {"error": errors[-1]})
            report = _report_error(state.run_id, messages, errors, events)
            return report, None

        emit("worker.started", self.name, {})
        return self._run_loop(
            adapter=adapter,
            job=state.job,
            run_id=state.run_id,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            messages=messages,
            tool_calls=tool_calls,
            metrics=metrics,
            errors=errors,
            iteration=iteration,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    async def aresume(
        self,
        *,
        adapter: BaseModelAdapter,
        state: RunState,
        decision_or_input: Any,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        pending = state.pending_action
        if pending is None:
            report = _build_report(
                state.run_id,
                "failed",
                None,
                None,
                None,
                state.messages,
                state.tool_calls,
                state.metrics,
                events,
                None,
                ["No pending action"],
            )
            return report, None

        messages = list(state.messages)
        tool_calls = list(state.tool_calls)
        iteration = state.iteration
        metrics = dict(state.metrics)
        errors: list[str] = []

        tool = self.toolbelt.resolve(pending.tool_call.name)
        if tool is None:
            result = ToolResult(error=f"Tool not found: {pending.tool_call.name}")
            emit(
                "tool.failed",
                pending.tool_call.name,
                {"tool_call_id": pending.tool_call.id, "error": result.error},
            )
            messages.append(tool_message(result, pending.tool_call))
            replace_tool_call(tool_calls, tool_call_with_result(pending.tool_call, result))
        else:
            if pending.type == "confirmation" and not decision_or_input:
                result = ToolResult(error="Tool execution declined")
                tool_result_message = tool_message(result, pending.tool_call)
                messages.append(tool_result_message)
                replace_tool_call(tool_calls, tool_call_with_result(pending.tool_call, result))
            else:
                call = pending.tool_call
                if pending.type == "user_input":
                    key = resume_argument_key(pending)
                    call = update_arguments(call, key, decision_or_input)
                emit("tool.started", tool.name, {"tool_call_id": call.id})
                result = await aexecute_tool(tool, call)
                if result.error:
                    emit("tool.failed", tool.name, {"tool_call_id": call.id, "error": result.error})
                else:
                    emit("tool.completed", tool.name, _tool_event_payload(call, result))
                tool_result_message = tool_message(result, call)
                messages.append(tool_result_message)
                replace_tool_call(tool_calls, tool_call_with_result(call, result))

        if not model_name:
            errors.append("Worker model not set")
            emit("worker.failed", self.name, {"error": errors[-1]})
            report = _report_error(state.run_id, messages, errors, events)
            return report, None

        emit("worker.started", self.name, {})
        return await self._arun_loop(
            adapter=adapter,
            job=state.job,
            run_id=state.run_id,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            messages=messages,
            tool_calls=tool_calls,
            metrics=metrics,
            errors=errors,
            iteration=iteration,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )



import json
import sqlite3
from typing import Any

from blackgeorge.memory.base import MemoryScope, MemoryStore
from blackgeorge.utils import utc_now


class SQLiteMemoryStore(MemoryStore):
    def __init__(self, path: str) -> None:
        self._path = path
        self._conn = sqlite3.connect(self._path)
        self._closed = False
        self._conn.execute(
            """
            CREATE TABLE IF NOT EXISTS memories (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                scope TEXT NOT NULL,
                key TEXT NOT NULL,
                value TEXT NOT NULL,
                created_at TEXT NOT NULL,
                updated_at TEXT NOT NULL,
                UNIQUE(scope, key)
            )
            """
        )
        self._conn.commit()

    def close(self) -> None:
        if not self._closed:
            self._conn.close()
            self._closed = True

    def __enter__(self) -> "SQLiteMemoryStore":
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()

    def write(self, key: str, value: Any, scope: MemoryScope) -> None:
        payload = json.dumps(value, ensure_ascii=True)
        now = utc_now().isoformat()
        self._conn.execute(
            """
            INSERT INTO memories (scope, key, value, created_at, updated_at)
            VALUES (?, ?, ?, ?, ?)
            ON CONFLICT(scope, key)
            DO UPDATE SET value = excluded.value, updated_at = excluded.updated_at
            """,
            (scope, key, payload, now, now),
        )
        self._conn.commit()

    def read(self, key: str, scope: MemoryScope) -> Any | None:
        cursor = self._conn.execute(
            "SELECT value FROM memories WHERE scope = ? AND key = ?",
            (scope, key),
        )
        row = cursor.fetchone()
        if not row:
            return None
        return json.loads(row[0])

    def search(self, query: str, scope: MemoryScope) -> list[tuple[str, Any]]:
        cursor = self._conn.execute(
            "SELECT key, value FROM memories WHERE scope = ?",
            (scope,),
        )
        matches: list[tuple[str, Any]] = []
        for key, value in cursor.fetchall():
            if query in key or query in value:
                matches.append((key, json.loads(value)))
        return matches

    def reset(self, scope: MemoryScope) -> None:
        self._conn.execute("DELETE FROM memories WHERE scope = ?", (scope,))
        self._conn.commit()



from typing import Any

import instructor


class InstructorClientPool:
    def __init__(self) -> None:
        self._sync: dict[str, Any] = {}
        self._async: dict[str, Any] = {}

    def get(self, model: str, async_client: bool) -> Any:
        if async_client:
            if model not in self._async:
                self._async[model] = instructor.from_provider(
                    f"litellm/{model}",
                    async_client=True,
                )
            return self._async[model]
        if model not in self._sync:
            self._sync[model] = instructor.from_provider(f"litellm/{model}")
        return self._sync[model]


instructor_clients = InstructorClientPool()



import json
from typing import Any

import chromadb
from chromadb.config import DEFAULT_DATABASE, DEFAULT_TENANT

from blackgeorge.memory.base import MemoryScope, MemoryStore
from blackgeorge.utils import utc_now

DEFAULT_CHUNK_SIZE = 8000
DEFAULT_CHUNK_OVERLAP = 200


def _chunk_text(
    text: str,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    overlap: int = DEFAULT_CHUNK_OVERLAP,
) -> list[str]:
    if len(text) <= chunk_size:
        return [text]
    if chunk_size <= 0:
        return [text]
    chunks: list[str] = []
    start = 0
    step = max(chunk_size - overlap, 1)
    while start < len(text):
        end = min(start + chunk_size, len(text))
        chunks.append(text[start:end])
        if end == len(text):
            break
        start += step
    return chunks


def _serialize_value(value: Any) -> str:
    if isinstance(value, str):
        return value
    return json.dumps(value, ensure_ascii=True, default=str)


def _deserialize_value(text: str) -> Any:
    try:
        return json.loads(text)
    except json.JSONDecodeError:
        return text


class VectorMemoryStore(MemoryStore):
    def __init__(
        self,
        path: str,
        collection_name: str = "memories",
        tenant: str = DEFAULT_TENANT,
        database: str = DEFAULT_DATABASE,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        chunk_overlap: int = DEFAULT_CHUNK_OVERLAP,
    ) -> None:
        self._path = path
        self._client = chromadb.PersistentClient(
            path=path,
            tenant=tenant,
            database=database,
        )
        self._collection_name = collection_name
        self._chunk_size = max(chunk_size, 1)
        self._chunk_overlap = max(chunk_overlap, 0)
        self._collection = self._client.get_or_create_collection(
            name=collection_name,
            metadata={"hnsw:space": "cosine"},
        )

    def _resolve_chunk_config(self, metadata: dict[str, Any]) -> tuple[int, int]:
        chunk_size_raw = metadata.get("chunk_size", self._chunk_size)
        chunk_overlap_raw = metadata.get("chunk_overlap", self._chunk_overlap)
        try:
            chunk_size = int(chunk_size_raw)
        except (TypeError, ValueError):
            chunk_size = self._chunk_size
        try:
            chunk_overlap = int(chunk_overlap_raw)
        except (TypeError, ValueError):
            chunk_overlap = self._chunk_overlap
        if chunk_size <= 0:
            chunk_size = self._chunk_size
        if chunk_overlap < 0:
            chunk_overlap = 0
        return chunk_size, chunk_overlap

    def write(self, key: str, value: Any, scope: MemoryScope) -> None:
        text = _serialize_value(value)
        chunks = _chunk_text(text, chunk_size=self._chunk_size, overlap=self._chunk_overlap)
        now = utc_now().isoformat()
        self._delete_by_key(key, scope)
        for i, chunk in enumerate(chunks):
            doc_id = f"{scope}:{key}:{i}"
            self._collection.add(
                ids=[doc_id],
                documents=[chunk],
                metadatas=[
                    {
                        "scope": scope,
                        "key": key,
                        "chunk_index": i,
                        "total_chunks": len(chunks),
                        "chunk_size": self._chunk_size,
                        "chunk_overlap": self._chunk_overlap,
                        "created_at": now,
                    }
                ],
            )

    def read(self, key: str, scope: MemoryScope) -> Any | None:
        results = self._collection.get(
            where={"$and": [{"scope": {"$eq": scope}}, {"key": {"$eq": key}}]},  # type: ignore[dict-item]
            include=["documents", "metadatas"],
        )
        if not results["ids"]:
            return None
        docs = results["documents"] or []
        metadatas = results["metadatas"] or []
        pairs: list[tuple[dict[str, Any], str]] = []
        for meta, doc in zip(metadatas, docs, strict=False):
            pairs.append((dict(meta), str(doc)))
        sorted_chunks = sorted(pairs, key=lambda x: int(x[0].get("chunk_index", 0)))
        chunk_meta = sorted_chunks[0][0] if sorted_chunks else {}
        chunk_size, chunk_overlap = self._resolve_chunk_config(chunk_meta)
        step = max(chunk_size - chunk_overlap, 1)
        overlap_size = chunk_size - step
        text_parts: list[str] = []
        for idx, (_, doc) in enumerate(sorted_chunks):
            if idx == 0:
                text_parts.append(doc)
            else:
                new_len = len(doc) - overlap_size
                if new_len > 0:
                    text_parts.append(doc[-new_len:])
        full_text = "".join(text_parts)
        return _deserialize_value(full_text)

    def search(
        self,
        query: str,
        scope: MemoryScope,
        top_k: int = 5,
    ) -> list[tuple[str, Any]]:
        results = self._collection.query(
            query_texts=[query],
            where={"scope": {"$eq": scope}},  # type: ignore[dict-item]
            n_results=top_k,
            include=["documents", "metadatas", "distances"],
        )
        if not results["ids"] or not results["ids"][0]:
            return []
        seen_keys: set[str] = set()
        matches: list[tuple[str, Any]] = []
        ids = results["ids"][0]
        metadatas = results["metadatas"][0] if results["metadatas"] else []
        for i, doc_id in enumerate(ids):
            meta = dict(metadatas[i]) if i < len(metadatas) else {}
            key = str(meta.get("key", doc_id))
            if key in seen_keys:
                continue
            seen_keys.add(key)
            value = self.read(key, scope)
            if value is not None:
                matches.append((key, value))
        return matches

    def reset(self, scope: MemoryScope) -> None:
        try:
            results = self._collection.get(
                where={"scope": {"$eq": scope}},  # type: ignore[dict-item]
                include=[],
            )
            if results["ids"]:
                self._collection.delete(ids=results["ids"])
        except Exception:
            pass

    def _delete_by_key(self, key: str, scope: MemoryScope) -> None:
        try:
            results = self._collection.get(
                where={"$and": [{"scope": {"$eq": scope}}, {"key": {"$eq": key}}]},  # type: ignore[dict-item]
                include=[],
            )
            if results["ids"]:
                self._collection.delete(ids=results["ids"])
        except Exception:
            pass



from collections.abc import AsyncIterator, Iterator
from typing import Any

from pydantic import BaseModel, ConfigDict

from blackgeorge.core.event import Event
from blackgeorge.core.job import Job
from blackgeorge.core.message import Message
from blackgeorge.core.report import Report
from blackgeorge.desk import Desk
from blackgeorge.store.session_store import SessionStore
from blackgeorge.store.sqlite_session_store import SQLiteSessionStore
from blackgeorge.utils import new_id
from blackgeorge.worker import Worker


class WorkerSession(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    session_id: str
    worker: Worker
    desk: Desk
    store: SessionStore

    @classmethod
    def start(
        cls,
        *,
        worker: Worker,
        desk: Desk,
        session_id: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> "WorkerSession":
        store = SQLiteSessionStore(desk.db_path)
        session_id = session_id or new_id()

        store.create_session(
            session_id=session_id,
            worker_name=worker.name,
            metadata=metadata or {},
        )

        return cls(
            session_id=session_id,
            worker=worker,
            desk=desk,
            store=store,
        )

    @classmethod
    def resume(
        cls,
        *,
        session_id: str,
        worker: Worker,
        desk: Desk,
    ) -> "WorkerSession | None":
        store = SQLiteSessionStore(desk.db_path)

        record = store.get_session(session_id)
        if record is None or record.worker_name != worker.name:
            return None

        return cls(
            session_id=session_id,
            worker=worker,
            desk=desk,
            store=store,
        )

    def run(self, user_input: Any, *, stream: bool = False, **job_kwargs: Any) -> Report:
        messages = self.store.get_messages(self.session_id)
        initial_count = len(messages) if messages else 0
        job = Job(
            input=user_input,
            initial_messages=messages if messages else None,
            **job_kwargs,
        )

        report = self.desk.run(self.worker, job, stream=stream)

        if report.status == "completed":
            new_messages = self._extract_conversation_messages(report)
            self.store.add_messages(self.session_id, new_messages[initial_count:])
            self.store.update_session(self.session_id)

        return report

    async def arun(self, user_input: Any, *, stream: bool = False, **job_kwargs: Any) -> Report:
        messages = self.store.get_messages(self.session_id)
        initial_count = len(messages) if messages else 0
        job = Job(
            input=user_input,
            initial_messages=messages if messages else None,
            **job_kwargs,
        )

        report = await self.desk.arun(self.worker, job, stream=stream)

        if report.status == "completed":
            new_messages = self._extract_conversation_messages(report)
            self.store.add_messages(self.session_id, new_messages[initial_count:])
            self.store.update_session(self.session_id)

        return report

    def history(self) -> list[Message]:
        return self.store.get_messages(self.session_id)

    def stream_run(self, user_input: Any, **job_kwargs: Any) -> Iterator[Event]:
        """Run worker and yield events from the completed report.

        Note: Events are yielded after the run completes, not in real-time.
        For true streaming during execution, use the event bus directly.
        """
        messages = self.store.get_messages(self.session_id)
        initial_count = len(messages) if messages else 0
        job = Job(
            input=user_input,
            initial_messages=messages if messages else None,
            **job_kwargs,
        )

        report = self.desk.run(self.worker, job, stream=True)

        if report.status == "completed":
            new_messages = self._extract_conversation_messages(report)
            self.store.add_messages(self.session_id, new_messages[initial_count:])
            self.store.update_session(self.session_id)

        yield from report.events

    async def astream_run(self, user_input: Any, **job_kwargs: Any) -> AsyncIterator[Event]:
        """Run worker and yield events from the completed report.

        Note: Events are yielded after the run completes, not in real-time.
        For true streaming during execution, use the event bus directly.
        """
        messages = self.store.get_messages(self.session_id)
        initial_count = len(messages) if messages else 0
        job = Job(
            input=user_input,
            initial_messages=messages if messages else None,
            **job_kwargs,
        )

        report = await self.desk.arun(self.worker, job, stream=True)

        if report.status == "completed":
            new_messages = self._extract_conversation_messages(report)
            self.store.add_messages(self.session_id, new_messages[initial_count:])
            self.store.update_session(self.session_id)

        for event in report.events:
            yield event

    def close(self) -> None:
        self.store.delete_session(self.session_id)

    def _extract_conversation_messages(self, report: Report) -> list[Message]:
        messages: list[Message] = []

        for message in report.messages:
            if (
                message.role in ("user", "assistant")
                or message.role == "tool"
                and message.tool_call_id
            ):
                msg_dict = message.model_dump()
                if message.role == "assistant":
                    msg_dict.pop("reasoning_content", None)
                messages.append(Message(**msg_dict))

        return messages
